
-- Clocks in data into fifo RAM and input pipeline
-- Calculates the census transform, when enough data was clocked in
-- Census calcualtion: = 1 if pixelcand >= centerpix. Index goes from left top (census(0)) to bottom right (census(maxindex))

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.allconstantsPack.ALL;	--contains windowlength and windowheight. From these values, the required Fifos are instantiated, the proper fifotopipe signals, pipetofifo, census length etc.
--use work.leftimagePack.ALL;

entity censuspipefiforam is
	generic(
		IMAGEWIDTH : integer := 1024;
		DATAWIDTH : integer := 8
	);
	port(
		clk : in std_logic;
		reset : in std_logic;
		datain : in unsigned(7 downto 0);
		--dataout : out unsigned(7 downto 0);		--Image pixel data drop out
		valid : out std_logic;
		targetrow : out unsigned(ImageHeightInBits - 1 downto 0);
		targetcol : out unsigned(ImageWidthInBits - 1 downto 0);
		censusout : out unsigned(NumberOfCensusValues - 1 downto 0)
	);
end censuspipefiforam;

architecture Behavioral of censuspipefiforam is

type pipeType IS ARRAY(NATURAL RANGE 0 TO WINDOWLENGTH - 1) OF unsigned(7 downto 0);

type procpipeType IS ARRAY(NATURAL RANGE 0 TO WINDOWHEIGHT - 1, 0 TO WINDOWLENGTH - 1) OF unsigned(DATAWIDTH - 1 DOWNTO 0);

--2D Pixel pipe at beginning and to work on
signal procPipe, procPipe_next : procpipeType;	--WINDOWHEIGHT times pipe		

--Signals between pipes and fifos:
type pipeToFifoType IS ARRAY(NATURAL RANGE 0 TO WINDOWHEIGHT - 2) OF unsigned(7 downto 0);
signal pipetofifo, pipetofifo_next : pipeToFifoType;

type fifoToPipeType IS ARRAY(NATURAL RANGE 0 TO WINDOWHEIGHT - 2) OF unsigned(7 downto 0);
signal fifotopipe : fifoToPipeType;

signal rowcsig, rowc_next : unsigned(ImageHeightInBits - 1 downto 0);
signal colcsig, colc_next : unsigned(ImageWidthInBits -1 downto 0);

signal validsig, valid_next : std_logic;
signal targetcolsig, targetcol_next : unsigned(ImageWidthInBits -1 downto 0);
signal targetrowsig, targetrow_next : unsigned(ImageHeightInBits - 1 downto 0);

signal census, census_next : unsigned(NumberOfCensusValues - 1 downto 0);

	COMPONENT fifowrapper
		GENERIC (
			DATAWIDTH : integer := DATAWIDTH;	--data range from 0 to (2^DATAWIDTH) - 1
			ADDRLEN : integer := IMAGEWIDTH - WindowLength		--FIFO delays data by ADDRLEN cycles. For 3x3, -2 must be added (nobody knows why)
		);
		PORT (
			clk : in std_logic;
			reset : in std_logic;
			wdata : in std_logic_vector((DATAWIDTH - 1) downto 0);
			rdatatop : out unsigned((DATAWIDTH - 1) downto 0)
		);
	end component;

BEGIN

	GENERATE_FIFO:
	FOR i IN 0 TO WINDOWHEIGHT - 2 GENERATE
		myfifox : fifowrapper PORT MAP(
			clk, reset, std_logic_vector(pipetofifo(i)), fifotopipe(i)
		);
	END GENERATE GENERATE_FIFO;
	
		
	clkProc : PROCESS(clk,reset)
	BEGIN
		IF (reset = '1') THEN
			rowcsig <= to_unsigned(1,ImageHeightInBits);--(OTHERS => '0');
			colcsig <= (OTHERS => '0');
			validsig <= '0';
			targetrowsig <= (OTHERS => '0');
			targetcolsig <= (OTHERS => '0');
			census <= (OTHERS => '0');
			procPipe <= (OTHERS => (OTHERS => (OTHERS => '0'))); 
			pipetofifo <= (OTHERS => (OTHERS => '0'));
			
		ELSIF (clk'event AND clk = '1') THEN
			rowcsig <= rowc_next;
			colcsig <= colc_next;
			validsig <= valid_next;
			census <= census_next;
			targetrowsig <= targetrow_next;
			targetcolsig <= targetcol_next;
			procPipe <= procPipe_next;
			pipetofifo <= pipetofifo_next;
			
		END IF; 
	END PROCESS clkProc;
	
	Comb_Proc : PROCESS(datain, procPipe, fifotopipe, colcsig, rowcsig, validsig)
	variable centerpix : unsigned(7 downto 0) := (OTHERS => '0');		--Size changes according to image data width!
	--variable indexcounter : unsigned(HammingBitSize - 1 downto 0) := (OTHERS => '0');
	variable indexcounter : integer range 0 to NumberOfCensusValues := 0; -- unsigned(HammingBitSize - 1 downto 0) := (OTHERS => '0');
	BEGIN
		
		--Shift the working pipe 
		FOR y in 0 TO WINDOWHEIGHT - 1 LOOP
			FOR x in 0 TO WINDOWLENGTH - 2 LOOP
				procpipe_next(y, x + 1) <= procPipe(y,x);
			END LOOP;
		END LOOP;
		
		--Connect the corret signals from and to fifo lines:
		FOR y in 0 TO WINDOWHEIGHT - 2 LOOP
			procPipe_next(y + 1,0) <= fifotopipe(y);
			pipetofifo_next(y) <= procPipe(y,2);
		END LOOP;
		
		--Data input and output to and from processing pipe
		procPipe_next(0,0) <= datain;
		--dataout <= procPipe(WINDOWHEIGHT - 1, WINDOWLENGTH - 1);
		
	--Counters:
		rowc_next <= rowcsig;
		colc_next <= colcsig;
		valid_next <= validsig;
		census_next <= (OTHERS => '0');
		
		IF colcsig < IMAGEWIDTH THEN
			colc_next <= colcsig + 1;
		ELSE 
			colc_next <= to_unsigned(1,ImageWidthInBits);
			--IF rowcsig < leftimageHeight THEN
			rowc_next <= rowcsig + 1;
			--ELSE
			--	rowc_next <= to_unsigned(1,ImageHeightInBits);
			--END IF;
		END IF;
		
		--Check, if enough data was clocked in and if so, calculate census transform:
		IF colcsig >= WINDOWLENGTH AND rowcsig >= WINDOWHEIGHT THEN
			targetcol_next <= colcsig - 1;
			targetrow_next <= rowcsig - 1;
			
			--centerpix := medpipe(1);
			centerpix := procPipe(WindowCenterY,WindowCenterX);
			--indexcounter := (OTHERS => '0');
			indexcounter := 0;
			
			--Run over the window but omit the center
			FOR y IN 0 TO WINDOWHEIGHT - 1 LOOP
				FOR x IN 0 TO WINDOWLENGTH - 1 LOOP
					IF NOT(y = WindowCenterY AND x = WindowCenterX) THEN
						IF procPipe(y,x) >= centerpix THEN 
							--census_next(to_integer(indexcounter)) <= '1';
							census_next(indexcounter) <= '1';
						END IF;
						indexcounter := indexcounter + 1;
					END IF;
				END LOOP;
			END LOOP;
			
			valid_next <= '1';
			
		ELSE
			valid_next <= '0';
			targetcol_next <= (OTHERS => '0');
			targetrow_next <= (OTHERS => '0');	
		END IF;
		
	END PROCESS Comb_Proc;
	
	
--Output ports:	
	valid <= validsig;
	censusout <= census;
	targetrow <= targetrowsig;
	targetcol <= targetcolsig;
	
end Behavioral;

