--Libraries
library ieee;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;

library work;
use work.allconstantsPack.ALL;

ENTITY censusselective IS
	PORT(
		clk : IN std_logic;
		reset : IN std_logic;
		leftin : IN unsigned(PixelBitWidth - 1 downto 0);
		rightin : IN unsigned(PixelBitWidth - 1 downto 0);
		expandin : IN expanddispType;		--expanded disparities, 1 TO 20
		hammingvectorout : OUT hammingvectorType	--hamming vector
	);
END censusselective;

ARCHITECTURE Behavioral OF censusselective IS

	COMPONENT censuspipefiforam 
		generic(
			IMAGEWIDTH : integer := ImageWidth;		--The census transform clockes in 2xIMAGEWIDTH + WINDOWLENGTH pixels before trasnform
			DATAWIDTH : integer := 8
		);
		PORT(
			clk : in std_logic;
			reset : in std_logic;
			datain : in unsigned(7 downto 0);
			--dataout : out unsigned(7 downto 0);		--Image pixel data drop out
			valid : out std_logic;
			targetrow : out unsigned(ImageHeightInBits - 1 downto 0);
			targetcol : out unsigned(ImageWidthInBits - 1 downto 0);
			censusout : out unsigned(NumberOfCensusValues - 1 downto 0)
		);
	END COMPONENT;

	--signal dataoutLsig, dataoutRsig : unsigned(7 downto 0);
	signal validRsig, validLsig : std_logic;
	signal censusLsig, censusRsig : unsigned(NumberOfCensusValues - 1 downto 0);
	
	TYPE censusvectorType IS ARRAY(natural range 1 TO D) OF unsigned(NumberOfCensusValues - 1 downto 0);	--For 9x9 CT, NumberOfCensusValues is 80!
	signal censusRpipe, censusRpipe_next : censusvectorType;
	
	signal colc, colc_next : unsigned(ImageWidthInBits - 1 downto 0);
	signal rowc, rowc_next : unsigned(ImageHeightInBits - 1 downto 0);
	
	signal targetrowL, targetrowR : unsigned(ImageHeightInBits - 1 downto 0);
	signal targetcolL, targetcolR : unsigned(ImageWidthInBits - 1 downto 0);
	
		--type hammingvectorType is ARRAY(1 TO EXPANDLEN) OF unsigned(NumberOfCensusValues - 1 downto 0);
	signal hammingvector, hammingvector_next : hammingvectorType;
	
BEGIN

	censusL : censuspipefiforam 
	PORT MAP(
		clk => clk,
		reset => reset,
		datain => leftin,
		valid => validLsig,
		targetrow => targetrowL,
		targetcol => targetcolL,
		censusout => censusLsig
	);
	
	censusR : censuspipefiforam 
	PORT MAP(
		clk => clk,
		reset => reset,
		datain => rightin,
		valid => validRsig,
		targetrow => targetrowR,
		targetcol => targetcolR,
		censusout => censusRsig
	);
	
	--Clocked process
	clkProc : PROCESS(clk, reset)
	begin
		IF reset = '1' then
			censusRpipe <= (OTHERS => (OTHERS => '0'));
			colc <= to_unsigned(1,ImageWidthInBits);
			rowc <= to_unsigned(1,ImageHeightInBits);	
			hammingvector <= (OTHERS => (OTHERS => '0'));
		ELSIF (clk'event AND clk = '1') THEN
			censusRpipe <= censusRpipe_next;
			colc <= colc_next;
			rowc <= rowc_next;
			hammingvector <= hammingvector_next;
		END IF;
	END PROCESS clkProc;

	censusPipeProc : PROCESS(censusRsig, censusRpipe)
	BEGIN
		censusRpipe_next(D) <= censusRsig;
		censusRpipe_next(1 TO D-1) <= censusRpipe(2 TO D);
	END PROCESS censusPipeProc;
	
	counterProc : PROCESS(colc, rowc) 
	BEGIN
		IF colc < IMAGEWIDTH THEN
			colc_next <= colc + 1;
			rowc_next <= rowc;
		ELSE
			colc_next <= to_unsigned(1,ImageWidthInBits);
			IF rowc < IMAGEHEIGHT THEN
				rowc_next <= rowc + 1;
			ELSE 
				rowc_next <= to_unsigned(1,ImageHeightInBits);
			END IF;
		END IF;
	END PROCESS counterProc;
	
--Degenerated Adder structure:
	censusProc1 : PROCESS(censusLsig, censusRpipe, expandin, hammingvector)
	variable temp : unsigned(HammingBitSize - 1 DOWNTO 0) := (OTHERS => '0');
	variable valuetocheck : unsigned(NumberOfCensusValues - 1 downto 0) := (OTHERS => '0');
	BEGIN
				
		hammingvector_next <= hammingvector;	--Register operation
		--Loop over all possible indices to check
		FOR j IN 1 TO EXPANDLEN LOOP
			temp := (OTHERS => '0');	
			IF expandin(j) > 0 THEN
				valuetocheck := censusRpipe(to_integer(expandin(j)));		--Variable is possibly not required
				FOR a in 0 TO NumberOfCensusValues - 1 LOOP
					IF (censusLsig(a) XOR valuetocheck(a)) = '1' THEN		--ODER IF censusLout(a) \= valuetocheck(a) THEN
						temp := temp + 1;
					END IF;
				END LOOP;
				
			ELSE
				temp := (OTHERS => '0');	--If no valid index available,
				--final := (OTHERS => '1');	--If no valid index available
			END IF;
			
			hammingvector_next(j) <= temp;
			--hammingvector_next(j) <= final;
		END LOOP;
			
	END PROCESS censusProc1;

	hammingvectorout <= hammingvector;


 end Behavioral;
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
