--Libraries
library ieee;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;

library work;
use work.allconstantsPack.ALL;
use work.BramAsLutPack_16x2.ALL;

--Using BRAM as LUT

ENTITY censusselective IS
	PORT(
		clk : IN std_logic;
		reset : IN std_logic;
		leftin : IN unsigned(PixelBitWidth - 1 downto 0);
		rightin : IN unsigned(PixelBitWidth - 1 downto 0);
		expandin : IN expanddispType;		--expanded disparities, 1 TO 20
		hammingvectorout : OUT hammingvectorType	--hamming vector
	);
END censusselective;

ARCHITECTURE Behavioral OF censusselective IS

	COMPONENT censuspipefiforam 
		generic(
			IMAGEWIDTH : integer := ImageWidth;		--The census transform clockes in 2xIMAGEWIDTH + WINDOWLENGTH pixels before trasnform
			DATAWIDTH : integer := 8
		);
		PORT(
			clk : in std_logic;
			reset : in std_logic;
			datain : in unsigned(7 downto 0);
			--dataout : out unsigned(7 downto 0);		--Image pixel data drop out
			valid : out std_logic;
			targetrow : out unsigned(ImageHeightInBits - 1 downto 0);
			targetcol : out unsigned(ImageWidthInBits - 1 downto 0);
			censusout : out unsigned(NumberOfCensusValues - 1 downto 0)
		);
	END COMPONENT;
	
	COMPONENT bramaslut PORT(
		clk : in std_logic;
		we : in std_logic;		--write enable
		waddr : in std_logic_vector(BramLenInBits - 1 downto 0);	
		wdata : in std_logic_vector(DataBitSize - 1 downto 0);
		raddr : in std_logic_vector(BramLenInBits - 1 downto 0);
		rdata : out unsigned(DataBitSize - 1 downto 0)
	);
	END COMPONENT;
	
	signal validRsig, validLsig : std_logic;
	signal censusLsig, censusRsig : unsigned(NumberOfCensusValues - 1 downto 0);
	
	TYPE censusvectorType IS ARRAY(natural range 1 TO D) OF unsigned(NumberOfCensusValues - 1 downto 0);	--For 9x9 CT, NumberOfCensusValues is 80!
	signal censusRpipe, censusRpipe_next : censusvectorType;
	
	signal colc, colc_next : unsigned(ImageWidthInBits - 1 downto 0);
	signal rowc, rowc_next : unsigned(ImageHeightInBits - 1 downto 0);
	
	signal targetrowL, targetrowR : unsigned(ImageHeightInBits - 1 downto 0);
	signal targetcolL, targetcolR : unsigned(ImageWidthInBits - 1 downto 0);
	
		--type hammingvectorType is ARRAY(1 TO EXPANDLEN) OF unsigned(NumberOfCensusValues - 1 downto 0);
	signal hammingvector, hammingvector_next : hammingvectorType;
	
	type bramresType IS ARRAY(natural range 1 TO EXPANDLEN, natural range 1 TO (NumberOfCensusValues/2)) OF unsigned(1 downto 0);
	signal bramres : bramresType;

	type bramaddrType IS ARRAY (natural range 1 TO EXPANDLEN, natural range 1 TO (NumberOfCensusValues/2)) OF unsigned(3 downto 0);
	signal bramaddr : bramaddrType;
	
BEGIN

	censusL : censuspipefiforam 
	PORT MAP(
		clk => clk,
		reset => reset,
		datain => leftin,
		valid => validLsig,
		targetrow => targetrowL,
		targetcol => targetcolL,
		censusout => censusLsig
	);
	
	censusR : censuspipefiforam 
	PORT MAP(
		clk => clk,
		reset => reset,
		datain => rightin,
		valid => validRsig,
		targetrow => targetrowR,
		targetcol => targetcolR,
		censusout => censusRsig
	);
	
	GENERATE_BRAM1:
	FOR j in 1 TO EXPANDLEN GENERATE
	BEGIN
		GENERATE_BRAM2:
		FOR i in 1 TO (NumberOfCensusValues/2) GENERATE
		BEGIN
			myBram : bramaslut PORT MAP(
				clk => clk,
				we => '0',
				waddr => (OTHERS => '0'), 
				wdata => (OTHERS => '0'), 
				raddr => std_logic_vector(bramaddr(j,i)),
				rdata => bramres(j,i)
			);
		END GENERATE GENERATE_BRAM2;
	END GENERATE GENERATE_BRAM1;
	
	
	
	--Clocked process
	clkProc : PROCESS(clk, reset)
	begin
		IF reset = '1' then
			censusRpipe <= (OTHERS => (OTHERS => '0'));
			colc <= to_unsigned(1,ImageWidthInBits);
			rowc <= to_unsigned(1,ImageHeightInBits);	
			hammingvector <= (OTHERS => (OTHERS => '0'));
		ELSIF (clk'event AND clk = '1') THEN
			censusRpipe <= censusRpipe_next;
			colc <= colc_next;
			rowc <= rowc_next;
			hammingvector <= hammingvector_next;
		END IF;
	END PROCESS clkProc;

	censusPipeProc : PROCESS(censusRsig, censusRpipe)
	BEGIN
		censusRpipe_next(D) <= censusRsig;
		censusRpipe_next(1 TO D-1) <= censusRpipe(2 TO D);
	END PROCESS censusPipeProc;
	
	counterProc : PROCESS(colc, rowc) 
	BEGIN
		IF colc < IMAGEWIDTH THEN
			colc_next <= colc + 1;
			rowc_next <= rowc;
		ELSE
			colc_next <= to_unsigned(1,ImageWidthInBits);
			IF rowc < IMAGEHEIGHT THEN
				rowc_next <= rowc + 1;
			ELSE 
				rowc_next <= to_unsigned(1,ImageHeightInBits);
			END IF;
		END IF;
	END PROCESS counterProc;
	
	
	censusProc3 : PROCESS(censusLsig, censusRpipe, expandin, hammingvector, bramres)
		
	--expandin: Array(1 .. Expandlen) of unsigned(HammingBitsize - 1 downto 0) -> Array(1 TO 20) OF unsigned(6 downto 0)
	--valueTocheck1 := rightpipe(expandin(1))
	--valueTocheck2 := rightpipe(expandin(2))
	--valueTocheck3 := rightpipe(expandin(3))
	--..
	--valueTocheckN := rightpipe(expandin(N))	N=Expandlen
	
	--valueToCheck : unsigned(80 downto 0)
	--tmp := valuuTocheck(i) & leftCensus(i) & valuetocheck(i+1) & leftcensus(i+1)
	
	type resvectorType2 IS ARRAY(NATURAL RANGE 1 TO 20) OF unsigned(2 DOWNTO 0);
	type resvectorType3 IS ARRAY(NATURAL RANGE 1 TO 10) OF unsigned(3 DOWNTO 0);
	type resvectorType4 IS ARRAY(NATURAL RANGE 1 TO 5) OF unsigned(4 DOWNTO 0);
	type resvectorType5 IS ARRAY(NATURAL RANGE 1 TO 3) OF unsigned(5 DOWNTO 0);
	type resvectorType6 IS ARRAY(NATURAL RANGE 1 TO 2) OF unsigned(6 DOWNTO 0);
	
	--variable res1 : resvectorType1;
	variable res2 : resvectorType2;
	variable res3 : resvectorType3;
	variable res4 : resvectorType4;
	variable res5 : resvectorType5;
	variable res6 : resvectorType6;
	
	variable tmp : unsigned(3 downto 0) := "0000";
	variable tmp2 : unsigned(5 downto 0) := "000000";
	variable valuetocheck : unsigned(NumberOfCensusValues - 1 downto 0) := (OTHERS => '0');
	variable final : unsigned(HammingBitSize - 1 downto 0);		--HammingBitSize 
	
	BEGIN
	
		hammingvector_next <= hammingvector;
		
		FOR j IN 1 TO EXPANDLEN LOOP
			final := (OTHERS => '0');
			IF expandin(j) > 0 THEN
				valuetocheck := censusRpipe(to_integer(expandin(j)));
	
				--Replace XORs or 16x2 ROMs with BRAM-LUTs to look up for 16 values the result, which is a 2 bit value:
				FOR i IN 0 TO (NumberOfCensusValues / 2) - 1 LOOP
					bramaddr(j,i+1) <= valuetocheck(i*2) & censusLsig(i*2) & valuetocheck((i*2) + 1) & censusLsig((i * 2) + 1);
				END LOOP;
	
				FOR i IN 1 TO (NumberOfCensusValues / 4) LOOP
					--tmp := res1((i*2)-1) & res1(i*2);
					tmp := bramres(j,(i*2)-1) & bramres(j,i*2);
					
					CASE tmp IS
						WHEN "0000" => res2(i) := "000";
						WHEN "0001" => res2(i) := "001";
						WHEN "0010" => res2(i) := "010";
						WHEN "0011" => res2(i) := "011";
						
						WHEN "0100" => res2(i) := "001";
						WHEN "0101" => res2(i) := "010";
						WHEN "0110" => res2(i) := "011";
						WHEN "0111" => res2(i) := "100";
						
						WHEN "1000" => res2(i) := "010";
						WHEN "1001" => res2(i) := "011";
						WHEN "1010" => res2(i) := "100";
						WHEN "1011" => res2(i) := "101";
						
						WHEN "1100" => res2(i) := "011";
						WHEN "1101" => res2(i) := "100";
						WHEN "1110" => res2(i) := "101";
						WHEN "1111" => res2(i) := "110";
						
						WHEN OTHERS => res2(i) := "000";
					END CASE;
				END LOOP;
				
				-- --next stage:
				-- res3(1) := ('0' & res2(1)) + ('0' & res2(2));	
				-- res3(2) := ('0' & res2(3)) + ('0' & res2(4));
				-- res3(3) := ('0' & res2(5)) + ('0' & res2(6));	
				-- res3(4) := ('0' & res2(7)) + ('0' & res2(8));
				
				-- res3(5) := ('0' & res2(9)) + ('0' & res2(10));	
				-- res3(6) := ('0' & res2(11)) + ('0' & res2(12));
				-- res3(7) := ('0' & res2(13)) + ('0' & res2(14));	
				-- res3(8) := ('0' & res2(15)) + ('0' & res2(16));
				
				-- --Third stage, add two 3-Bit values:
				FOR i IN 1 TO (NumberOfCensusValues / 8) LOOP
					tmp2 := res2((i*2) - 1) & res2(i*2);	--Concat two 3-Bit values to a 6 Bit value
					
					CASE tmp2 IS
						WHEN "000000" => res3(i) := "0000";		--1
						WHEN "000001" => res3(i) := "0001";
						WHEN "000010" => res3(i) := "0010";
						WHEN "000011" => res3(i) := "0011";
						
						WHEN "000100" => res3(i) := "0100";		--2
						WHEN "000101" => res3(i) := "0101";
						WHEN "000110" => res3(i) := "0110";
						WHEN "000111" => res3(i) := "0111";
						
						WHEN "001000" => res3(i) := "0001";		--3
						WHEN "001001" => res3(i) := "0010";
						WHEN "001010" => res3(i) := "0011";
						WHEN "001011" => res3(i) := "0100";
						
						WHEN "001100" => res3(i) := "0101";		--4
						WHEN "001101" => res3(i) := "0110";
						WHEN "001110" => res3(i) := "0111";
						WHEN "001111" => res3(i) := "1000";
						
						WHEN "010000" => res3(i) := "0010";		--5
						WHEN "010001" => res3(i) := "0011";
						WHEN "010010" => res3(i) := "0100";
						WHEN "010011" => res3(i) := "0101";
						
						WHEN "010100" => res3(i) := "0110";		--6
						WHEN "010101" => res3(i) := "0111";
						WHEN "010110" => res3(i) := "1000";
						WHEN "010111" => res3(i) := "1001";
						
						WHEN "011000" => res3(i) := "0011";		--7
						WHEN "011001" => res3(i) := "0100";
						WHEN "011010" => res3(i) := "0101";
						WHEN "011011" => res3(i) := "0110";
						
						WHEN "011100" => res3(i) := "0111";		--8
						WHEN "011101" => res3(i) := "1000";
						WHEN "011110" => res3(i) := "1001";
						WHEN "011111" => res3(i) := "1010";
						
						WHEN "100000" => res3(i) := "0100";		--9
						WHEN "100001" => res3(i) := "0101";
						WHEN "100010" => res3(i) := "0110";
						WHEN "100011" => res3(i) := "0111";
						
						WHEN "100100" => res3(i) := "1000";		--10
						WHEN "100101" => res3(i) := "1001";
						WHEN "100110" => res3(i) := "1010";
						WHEN "100111" => res3(i) := "1011";
						
						WHEN "101000" => res3(i) := "0101";		--11
						WHEN "101001" => res3(i) := "0110";
						WHEN "101010" => res3(i) := "0111";
						WHEN "101011" => res3(i) := "1000";
						
						WHEN "101100" => res3(i) := "1001";		--12
						WHEN "101101" => res3(i) := "1010";
						WHEN "101110" => res3(i) := "1011";
						WHEN "101111" => res3(i) := "1100";
						
						WHEN "110000" => res3(i) := "0110";		--13
						WHEN "110001" => res3(i) := "0111";
						WHEN "110010" => res3(i) := "1000";
						WHEN "110011" => res3(i) := "1001";
						
						WHEN "110100" => res3(i) := "1010";		--14
						WHEN "110101" => res3(i) := "1011";
						WHEN "110110" => res3(i) := "1100";
						WHEN "110111" => res3(i) := "1101";
						
						WHEN "111000" => res3(i) := "0111";		--15
						WHEN "111001" => res3(i) := "1000";
						WHEN "111010" => res3(i) := "1001";
						WHEN "111011" => res3(i) := "1010";
						
						WHEN "111100" => res3(i) := "1011";		--16
						WHEN "111101" => res3(i) := "1100";
						WHEN "111110" => res3(i) := "1101";
						WHEN "111111" => res3(i) := "1110";
						
						WHEN OTHERS => res3(i) := "0000";
					END CASE;
				END LOOP;
				
				
				res4(1) := ('0' & res3(1)) + ('0' & res3(2));
				res4(2) := ('0' & res3(3)) + ('0' & res3(4));
				res4(3) := ('0' & res3(5)) + ('0' & res3(6));
				res4(4) := ('0' & res3(7)) + ('0' & res3(8));
				res4(5) := ('0' & res3(9)) + ('0' & res3(10));

				--Stage 5:
				res5(1) := ('0' & res4(1)) + ('0' & res4(2));
				res5(2) := ('0' & res4(3)) + ('0' & res4(4));
				res5(3) := ('0' & res4(5));

				--Stage 6:
				res6(1) := ('0' & res5(1)) + ('0' & res5(2));
				res6(2) := ('0' & res5(3));
				
				final := res6(1) + res6(2);	
			ELSE
				final := (OTHERS => '0');
				bramaddr <= (OTHERS => (OTHERS => (OTHERS => '0')));
			END IF;
			
			hammingvector_next(j) <= final;
		END LOOP;
	
	END PROCESS;
		
	--censusRout <= censusRpipe;
	--censusLout <= censusLsig;
	
	hammingvectorout <= hammingvector;


 end Behavioral;
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
